package ifneeded awthemes 10.4.0 \
    [list source [file join $dir awthemes.tcl]]
package ifneeded colorutils 4.8 \
    [list source [file join $dir colorutils.tcl]]
package ifneeded awarc 1.6.1 \
    [list source [file join $dir awarc.tcl]]
package ifneeded ttk::theme::awarc 1.6.1 \
    [list source [file join $dir awarc.tcl]]
package ifneeded awblack 7.8.1 \
    [list source [file join $dir awblack.tcl]]
package ifneeded ttk::theme::awblack 7.8.1 \
    [list source [file join $dir awblack.tcl]]
package ifneeded awbreeze 1.9.1 \
    [list source [file join $dir awbreeze.tcl]]
package ifneeded ttk::theme::awbreeze 1.9.1 \
    [list source [file join $dir awbreeze.tcl]]
package ifneeded awbreezedark 1.0.1 \
    [list source [file join $dir awbreezedark.tcl]]
package ifneeded ttk::theme::awbreezedark 1.0.1 \
    [list source [file join $dir awbreezedark.tcl]]
package ifneeded awclearlooks 1.3.1 \
    [list source [file join $dir awclearlooks.tcl]]
package ifneeded ttk::theme::awclearlooks 1.3.1 \
    [list source [file join $dir awclearlooks.tcl]]
package ifneeded awdark 7.12 \
    [list source [file join $dir awdark.tcl]]
package ifneeded ttk::theme::awdark 7.12 \
    [list source [file join $dir awdark.tcl]]
package ifneeded awlight 7.10 \
    [list source [file join $dir awlight.tcl]]
package ifneeded ttk::theme::awlight 7.10 \
    [list source [file join $dir awlight.tcl]]
package ifneeded awtemplate 1.5.1 \
    [list source [file join $dir awtemplate.tcl]]
package ifneeded ttk::theme::awtemplate 1.5.1 \
    [list source [file join $dir awtemplate.tcl]]
package ifneeded awwinxpblue 7.9.1 \
    [list source [file join $dir awwinxpblue.tcl]]
package ifneeded ttk::theme::awwinxpblue 7.9.1 \
    [list source [file join $dir awwinxpblue.tcl]]
